"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProtectCamera = exports.PROTECT_SWITCH_TRIGGER = exports.PROTECT_SWITCH_MOTION = void 0;
const protect_accessory_1 = require("./protect-accessory");
const protect_stream_1 = require("./protect-stream");
// Manage our switch types.
exports.PROTECT_SWITCH_MOTION = "MotionSensorSwitch";
exports.PROTECT_SWITCH_TRIGGER = "MotionSensorTrigger";
class ProtectCamera extends protect_accessory_1.ProtectAccessory {
    // Configure a camera accessory for HomeKit.
    async configureDevice() {
        var _a;
        this.isVideoConfigured = false;
        // Save the camera object before we wipeout the context.
        const camera = this.accessory.context.camera;
        // Default motion detection support to on.
        let detectMotion = true;
        // Save the motion sensor switch state before we wipeout the context.
        if (this.accessory.context.detectMotion !== undefined) {
            detectMotion = this.accessory.context.detectMotion;
        }
        // Clean out the context object in case it's been polluted somehow.
        this.accessory.context = {};
        this.accessory.context.camera = camera;
        this.accessory.context.nvr = (_a = this.nvr.nvrApi.bootstrap) === null || _a === void 0 ? void 0 : _a.nvr.mac;
        this.accessory.context.detectMotion = detectMotion;
        // Clear out any switches on this camera so we can start fresh.
        let switchService;
        while ((switchService = this.accessory.getService(this.hap.Service.Switch))) {
            this.accessory.removeService(switchService);
        }
        // Configure accessory information.
        this.configureInfo();
        // Configure MQTT services.
        this.configureMqtt();
        // Configure the motion sensor.
        this.configureMotionSensor();
        this.configureMotionSwitch();
        this.configureMotionTrigger();
        // Configure two-way audio support and our video stream...and we're done.
        this.configureTwoWayAudio();
        return await this.configureVideoStream();
    }
    // Configure the camera device information for HomeKit.
    configureInfo() {
        var _a, _b, _c, _d, _e;
        const accessory = this.accessory;
        const camera = accessory.context.camera;
        const hap = this.hap;
        // Update the manufacturer information for this camera.
        (_a = accessory
            .getService(hap.Service.AccessoryInformation)) === null || _a === void 0 ? void 0 : _a.getCharacteristic(hap.Characteristic.Manufacturer).updateValue("Ubiquiti Networks");
        // Update the model information for this camera.
        (_b = accessory
            .getService(hap.Service.AccessoryInformation)) === null || _b === void 0 ? void 0 : _b.getCharacteristic(hap.Characteristic.Model).updateValue(camera.type);
        // Update the serial number for this camera.
        (_c = accessory
            .getService(hap.Service.AccessoryInformation)) === null || _c === void 0 ? void 0 : _c.getCharacteristic(hap.Characteristic.SerialNumber).updateValue(camera.mac);
        // Update the hardware revision for this camera.
        (_d = accessory
            .getService(hap.Service.AccessoryInformation)) === null || _d === void 0 ? void 0 : _d.getCharacteristic(hap.Characteristic.HardwareRevision).updateValue(camera.hardwareRevision);
        // Update the firmware revision for this camera.
        (_e = accessory
            .getService(hap.Service.AccessoryInformation)) === null || _e === void 0 ? void 0 : _e.getCharacteristic(hap.Characteristic.FirmwareRevision).updateValue(camera.firmwareVersion);
        return true;
    }
    // Configure the camera motion sensor for HomeKit.
    configureMotionSensor() {
        var _a;
        const accessory = this.accessory;
        const hap = this.hap;
        // Clear out any previous motion sensor service.
        let motionService = accessory.getService(hap.Service.MotionSensor);
        if (motionService) {
            accessory.removeService(motionService);
        }
        // Have we disabled motion sensors?
        if (!((_a = this.nvr) === null || _a === void 0 ? void 0 : _a.optionEnabled(accessory.context.camera, "MotionSensor"))) {
            this.log("%s: Disabling motion sensor.", this.name());
            return false;
        }
        // Add the motion sensor to the camera.
        motionService = new hap.Service.MotionSensor(accessory.displayName);
        accessory.addService(motionService);
        return true;
    }
    // Configure a switch to easily activate or deactivate motion sensor detection for HomeKit.
    configureMotionSwitch() {
        var _a, _b, _c;
        // Clear out any previous switch service.
        let switchService = this.accessory.getServiceById(this.hap.Service.Switch, exports.PROTECT_SWITCH_MOTION);
        if (switchService) {
            this.accessory.removeService(switchService);
        }
        // Have we disabled motion sensors or the motion switch?
        if (!((_a = this.nvr) === null || _a === void 0 ? void 0 : _a.optionEnabled(this.accessory.context.camera, "MotionSensor")) ||
            !((_b = this.nvr) === null || _b === void 0 ? void 0 : _b.optionEnabled(this.accessory.context.camera, "MotionSwitch"))) {
            this.log("%s: Disabling motion sensor switch.", this.name());
            // If we disable the switch, make sure we fully reset it's state.
            this.accessory.context.detectMotion = true;
            return false;
        }
        // Add the switch to the camera.
        switchService = new this.hap.Service.Switch(this.accessory.displayName + " Motion Events", exports.PROTECT_SWITCH_MOTION);
        // Activate or deactivate motion detection.
        (_c = this.accessory.addService(switchService)
            .getCharacteristic(this.hap.Characteristic.On)) === null || _c === void 0 ? void 0 : _c.on("get" /* GET */, (callback) => {
            callback(null, this.accessory.context.detectMotion === true);
        }).on("set" /* SET */, (value, callback) => {
            if (this.accessory.context.detectMotion !== value) {
                this.log("%s: Motion detection %s.", this.name(), (value === true) ? "enabled" : "disabled");
            }
            this.accessory.context.detectMotion = value === true;
            callback(null);
        }).updateValue(this.accessory.context.detectMotion);
        return true;
    }
    // Configure a switch to manually trigger a motion sensor event for HomeKit.
    configureMotionTrigger() {
        var _a, _b, _c;
        // Clear out any previous switch service.
        let triggerService = this.accessory.getServiceById(this.hap.Service.Switch, exports.PROTECT_SWITCH_TRIGGER);
        if (triggerService) {
            this.accessory.removeService(triggerService);
        }
        // Motion triggers are disabled by default and primarily exist for automation purposes.
        if (!((_a = this.nvr) === null || _a === void 0 ? void 0 : _a.optionEnabled(this.accessory.context.camera, "MotionSensor")) ||
            !((_b = this.nvr) === null || _b === void 0 ? void 0 : _b.optionEnabled(this.accessory.context.camera, "MotionTrigger", false))) {
            return false;
        }
        // Add the switch to the camera.
        triggerService = new this.hap.Service.Switch(this.accessory.displayName + " Motion Trigger", exports.PROTECT_SWITCH_TRIGGER);
        const motionService = this.accessory.getService(this.hap.Service.MotionSensor);
        const switchService = this.accessory.getServiceById(this.hap.Service.Switch, exports.PROTECT_SWITCH_MOTION);
        // Activate or deactivate motion detection.
        (_c = this.accessory.addService(triggerService)
            .getCharacteristic(this.hap.Characteristic.On)) === null || _c === void 0 ? void 0 : _c.on("get" /* GET */, (callback) => {
            callback(null, motionService === null || motionService === void 0 ? void 0 : motionService.getCharacteristic(this.hap.Characteristic.MotionDetected).value);
        }).on("set" /* SET */, (value, callback) => {
            if (value) {
                // Check to see if motion events are disabled.
                if (switchService && !switchService.getCharacteristic(this.hap.Characteristic.On).value) {
                    setTimeout(() => {
                        triggerService === null || triggerService === void 0 ? void 0 : triggerService.getCharacteristic(this.hap.Characteristic.On).updateValue(false);
                    }, 50);
                }
                else {
                    // Trigger the motion event.
                    this.nvr.motionEventHandler(this.accessory, Date.now());
                    this.log("%s: Motion event triggered.", this.name());
                }
            }
            else {
                // If the motion sensor is still on, we should be as well.
                if (motionService === null || motionService === void 0 ? void 0 : motionService.getCharacteristic(this.hap.Characteristic.MotionDetected).value) {
                    setTimeout(() => {
                        triggerService === null || triggerService === void 0 ? void 0 : triggerService.getCharacteristic(this.hap.Characteristic.On).updateValue(true);
                    }, 50);
                }
            }
            callback(null);
        }).updateValue(false);
        this.log("%s: Enabling motion sensor automation trigger.", this.name());
        return true;
    }
    // Configure two-way audio support for HomeKit.
    configureTwoWayAudio() {
        var _a, _b;
        // Identify twoway-capable devices.
        switch ((_a = this.accessory.context.camera) === null || _a === void 0 ? void 0 : _a.type) {
            case "UVC G3 Micro":
            case "UVC G4 Doorbell":
                // Enabled by default unless disabled by the user.
                this.twoWayAudio = (_b = this.nvr) === null || _b === void 0 ? void 0 : _b.optionEnabled(this.accessory.context.camera, "TwoWayAudio");
                break;
            default:
                this.twoWayAudio = false;
                break;
        }
        return true;
    }
    // Configure a camera accessory for HomeKit.
    async configureVideoStream() {
        var _a, _b;
        const bootstrap = this.nvr.nvrApi.bootstrap;
        const nvr = this.nvr;
        const nvrApi = this.nvr.nvrApi;
        // No channels exist on this camera or we don't have access to the bootstrap configuration.
        if (!((_a = this.accessory.context.camera) === null || _a === void 0 ? void 0 : _a.channels) || !bootstrap) {
            return false;
        }
        const camera = (_b = await nvrApi.enableRtsp(this.accessory.context.camera)) !== null && _b !== void 0 ? _b : this.accessory.context.camera;
        let forceQuality = "";
        let newCameraQuality = "";
        let newCameraUrl = "";
        if (nvr.optionEnabled(camera, "StreamOnly.Low", false)) {
            forceQuality = "Low";
        }
        else if (nvr.optionEnabled(camera, "StreamOnly.Medium", false)) {
            forceQuality = "Medium";
        }
        else if (nvr.optionEnabled(camera, "StreamOnly.High", false)) {
            forceQuality = "High";
        }
        // Filter the stream the user has explicitly set. If we can't find the stream quality the
        // user requests, we fail.
        if (forceQuality) {
            const foundChannel = camera.channels.find(channel => {
                // No RTSP channel here.
                if (!channel.isRtspEnabled) {
                    return false;
                }
                return channel.name === forceQuality;
            });
            if (foundChannel) {
                newCameraQuality = foundChannel.name;
                newCameraUrl = foundChannel.rtspAlias;
            }
        }
        else {
            // Our defaults may seem counterintuitive in some cases but here's the rationale:
            // The G4-series of cameras, particularly the G4 Pro and the G4 Bullet push out
            // 3840x2160 and 2688x1520, respectively, if you use the "High" stream setting --
            // that's a lot of data! HomeKit can only handle up to 1920x1080 (1080p) streams -
            // trying to push anything larger through to HomeKit is pointless, at least at this
            // time. Instead, to increase our chances of a smooth and quick playback experience,
            // we're going to set defaults that make sense within the context of HomeKit.
            let channelPriority;
            switch (camera.type) {
                // Handle very high-resolution cameras with more sane HomeKit defaults.
                case "UVC G4 Pro":
                case "UVC G4 Bullet":
                    channelPriority = ["Medium", "Low", "High"];
                    break;
                default:
                    channelPriority = ["High", "Medium", "Low"];
                    break;
            }
            // Iterate.
            for (const quality of channelPriority) {
                const foundChannel = camera.channels.find(channel => {
                    // No RTSP channel here.
                    if (!channel.isRtspEnabled) {
                        return false;
                    }
                    // Honor the user's quality biases.
                    if (!nvr.optionEnabled(camera, "Stream." + quality)) {
                        return false;
                    }
                    return channel.name === quality;
                });
                if (foundChannel) {
                    newCameraQuality = foundChannel.name;
                    newCameraUrl = foundChannel.rtspAlias;
                    break;
                }
            }
        }
        // No RTSP stream is available.
        if (!newCameraUrl) {
            // Notify only if this is a new change.
            if (!this.isVideoConfigured || this.cameraUrl) {
                this.log("%s: No RTSP stream has been configured for this camera. %s", this.name(), "Enable an RTSP stream in the UniFi Protect webUI to resolve this issue or " +
                    "assign the Administrator role to the user configured for this plugin to allow it to automatically configure itself.");
            }
        }
        else {
            // Set the selected quality.
            newCameraUrl = "rtsp://" + bootstrap.nvr.host + ":" + bootstrap.nvr.ports.rtsp.toString() + "/" + newCameraUrl;
            if (this.cameraUrl !== newCameraUrl) {
                this.log("%s: Stream quality configured: %s.", this.name(), newCameraQuality);
            }
        }
        // Set the video stream and shapshot URLs.
        this.cameraUrl = newCameraUrl;
        this.snapshotUrl = nvrApi.camerasUrl() + "/" + camera.id + "/snapshot";
        // Configure the video stream and inform HomeKit about it, if it's our first time.
        if (!this.isVideoConfigured) {
            this.isVideoConfigured = true;
            this.stream = new protect_stream_1.ProtectStreamingDelegate(this);
            this.accessory.configureController(this.stream.controller);
        }
        return true;
    }
    // Configure MQTT capabilities of this camera.
    configureMqtt() {
        var _a, _b, _c;
        const bootstrap = this.nvr.nvrApi.bootstrap;
        const camera = this.accessory.context.camera;
        // Trigger a motion event in MQTT, if requested to do so.
        (_a = this.nvr.mqtt) === null || _a === void 0 ? void 0 : _a.subscribe(this.accessory, "motion/trigger", (message) => {
            const value = message.toString();
            // When we get the right message, we trigger the motion event.
            if ((value === null || value === void 0 ? void 0 : value.toLowerCase()) !== "true") {
                return;
            }
            // Trigger the motion event.
            this.nvr.motionEventHandler(this.accessory, Date.now());
            this.log("%s: Motion event triggered via MQTT.", this.name());
        });
        // Return the RTSP URLs when requested.
        (_b = this.nvr.mqtt) === null || _b === void 0 ? void 0 : _b.subscribe(this.accessory, "rtsp/get", (message) => {
            var _a;
            const value = message.toString();
            // When we get the right message, we trigger the snapshot request.
            if ((value === null || value === void 0 ? void 0 : value.toLowerCase()) !== "true") {
                return;
            }
            const urlInfo = {};
            // Include the default URL we're using for the camera.
            if (this.cameraUrl) {
                urlInfo.Default = this.cameraUrl;
            }
            // Grab all the available RTSP channels.
            for (const channel of camera.channels) {
                if (!bootstrap || !channel.isRtspEnabled) {
                    continue;
                }
                urlInfo[channel.name] = "rtsp://" + bootstrap.nvr.host + ":" + bootstrap.nvr.ports.rtsp.toString() + "/" + channel.rtspAlias;
            }
            (_a = this.nvr.mqtt) === null || _a === void 0 ? void 0 : _a.publish(this.accessory, "rtsp", JSON.stringify(urlInfo));
            this.log("%s: RTSP information published via MQTT.", this.name());
        });
        // Trigger snapshots when requested.
        (_c = this.nvr.mqtt) === null || _c === void 0 ? void 0 : _c.subscribe(this.accessory, "snapshot/trigger", (message) => {
            var _a;
            const value = message.toString();
            // When we get the right message, we trigger the snapshot request.
            if ((value === null || value === void 0 ? void 0 : value.toLowerCase()) !== "true") {
                return;
            }
            void ((_a = this.stream) === null || _a === void 0 ? void 0 : _a.getSnapshot());
            this.log("%s: Snapshot triggered via MQTT.", this.name());
        });
        return true;
    }
    // Utility function for reserved identifiers for switches.
    isReservedName(name) {
        return name === undefined ? false :
            [exports.PROTECT_SWITCH_MOTION.toUpperCase(), exports.PROTECT_SWITCH_TRIGGER.toUpperCase()].includes(name.toUpperCase());
    }
}
exports.ProtectCamera = ProtectCamera;
//# sourceMappingURL=protect-camera.js.map