"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FfmpegProcess = void 0;
/* Copyright(C) 2017-2020, HJD (https://github.com/hjdhjd). All rights reserved.
 *
 * protect-ffmpeg.ts: FFmpeg capability validation and process control.
 *
 * This module is heavily inspired by the homebridge and homebridge-camera-ffmpeg source code and
 * borrows heavily from both. Thank you for your contributions to the HomeKit world.
 */
const dgram_1 = require("dgram");
const execa_1 = __importDefault(require("execa"));
class FfmpegProcess {
    constructor(delegate, sessionId, command, returnPort, callback) {
        this.debug = delegate.platform.debug.bind(this);
        this.delegate = delegate;
        this.log = delegate.platform.log;
        this.sessionId = sessionId;
        // Toggle FFmpeg logging, if configured.
        this.isVerbose = this.delegate.platform.verboseFfmpeg;
        if (this.isVerbose) {
            this.log("%s: ffmpeg command: %s %s", delegate.protectCamera.name(), delegate.videoProcessor, command.join(" "));
        }
        else {
            this.debug("%s: ffmpeg command: %s %s", delegate.protectCamera.name(), delegate.videoProcessor, command.join(" "));
        }
        // Create the return port for FFmpeg, if requested to do so. The only time we don't do this is when we're standing up
        // a two-way audio stream - in that case, the audio work is done through RtpSplitter and not here.
        if (returnPort) {
            this.createSocket(returnPort);
        }
        void this.startFfmpeg(command, callback);
    }
    // Create the port for FFmpeg to send data through.
    createSocket(portInfo) {
        const socket = dgram_1.createSocket(portInfo.addressVersion === "ipv6" ? "udp6" : "udp4");
        // Handle potential network errors.
        socket.on("error", (error) => {
            this.log.error("%s: Socket error: ", this.delegate.protectCamera.name(), error.name);
            this.delegate.stopStream(this.sessionId);
        });
        // Kill zombie video streams.
        socket.on("message", () => {
            if (this.timeout) {
                clearTimeout(this.timeout);
            }
            this.timeout = setTimeout(() => {
                this.log("%s: Device appears to be inactive for over 5 seconds. Stopping stream.", this.delegate.protectCamera.name());
                this.delegate.controller.forceStopStreamingSession(this.sessionId);
                this.delegate.stopStream(this.sessionId);
            }, 5000);
        });
        socket.bind(portInfo.port);
    }
    // Start our FFmpeg process.
    async startFfmpeg(ffmpegCommandLine, callback) {
        var _a, _b, _c, _d;
        // Track if we've started receiving data.
        let started = false;
        // Prepare the command line we want to execute.
        this.process = execa_1.default(this.delegate.videoProcessor, ffmpegCommandLine);
        // Handle errors on stdin.
        (_a = this.process.stdin) === null || _a === void 0 ? void 0 : _a.on("error", (error) => {
            if (!error.message.includes("EPIPE")) {
                this.log.error("%s: FFmpeg error: %s", this.delegate.protectCamera.name(), error.message);
            }
        });
        // Handle logging output that gets sent to stderr.
        (_b = this.process.stderr) === null || _b === void 0 ? void 0 : _b.on("data", (data) => {
            if (!started) {
                started = true;
                this.debug("%s: Received the first frame.", this.delegate.protectCamera.name());
                // Always remember to execute the callback once we're setup to let homebridge know we're streaming.
                if (callback) {
                    callback();
                }
            }
            // Debugging and additional logging, if requested.
            if (this.isVerbose || this.delegate.platform.debugMode) {
                data.toString().split(/\n/).forEach((line) => {
                    this.log(line);
                });
            }
        });
        try {
            // Execute the command line.
            await this.process;
        }
        catch (error) {
            // You might think this should be ExecaError, but ExecaError is a type, not a class, and instanceof
            // only operates on classes.
            if (!(error instanceof Error)) {
                this.log("Unknown error received while attempting to start FFmpeg: %s.", error);
                return;
            }
            // Recast our error object as an ExecaError.
            const execError = error;
            // Some utilities to streamline things.
            const logPrefix = this.delegate.protectCamera.name() + ": FFmpeg process ended ";
            const code = (_c = execError.exitCode) !== null && _c !== void 0 ? _c : null;
            const signal = (_d = execError.signal) !== null && _d !== void 0 ? _d : null;
            // We asked FFmpeg to stop.
            if (execError.isCanceled) {
                this.debug(logPrefix + "(Expected).");
                return;
            }
            // FFmpeg ended for another reason.
            const errorMessage = logPrefix + "(Error)." + (code === null ? "" : " Exit code: " + code.toString() + ".") + (signal === null ? "" : " Signal: " + signal + ".");
            this.log.error("%s: %s", this.delegate.protectCamera.name(), execError.message);
            // Stop the stream.
            this.delegate.stopStream(this.sessionId);
            // Temporarily increase logging verbosity.
            this.delegate.setVerboseFfmpeg();
            // Let homebridge know what happened and stop the stream if we've already started.
            if (!started && callback) {
                callback(new Error(errorMessage));
            }
            else {
                this.delegate.controller.forceStopStreamingSession(this.sessionId);
            }
        }
    }
    // Cleanup after we're done.
    stop() {
        // Cancel our process.
        this.process.cancel();
        // Kill our heartbeat monitoring.
        if (this.timeout) {
            clearTimeout(this.timeout);
        }
    }
    // Grab the standard input.
    getStdin() {
        return this.process.stdin;
    }
    // Grab the standard output.
    getStdout() {
        return this.process.stdout;
    }
    // Validate whether or not we have a specific codec available to us in FFmpeg.
    static async codecEnabled(videoProcessor, codec) {
        const output = await execa_1.default(videoProcessor, ["-codecs"]);
        return output.stdout.includes(codec);
    }
}
exports.FfmpegProcess = FfmpegProcess;
//# sourceMappingURL=protect-ffmpeg.js.map