"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RtpUtils = exports.RtpSplitter = void 0;
/* Copyright(C) 2017-2020, HJD (https://github.com/hjdhjd). All rights reserved.
 *
 * protect-rtp.ts: RTP-related utilities to slice and dice RTP streams.
 *
 * This module is heavily inspired by the homebridge and homebridge-camera-ffmpeg source code and
 * borrows heavily from both. Thank you for your contributions to the HomeKit world.
 */
const dgram_1 = require("dgram");
const get_port_1 = __importDefault(require("get-port"));
const settings_1 = require("./settings");
// What this function does is create two socket pipes to split traffic coming in from serverPort and
// pipe it out to returnAudioPort and twowayAudioPort.
//
// In order to support two-way audio, we send out the inbound RTP traffic from serverPort to twowayAudioPort.
//
// For the audio channel that supports the video stream we're sending to the user, we only want to send the RTCP
// information, because that channel is receiving all it's data (audio and video) from the video stream. RTCP
// provides quality and statistics information back to HomeKit.
//
// Credit to @dgreif and @brandawg93 who graciously shared their code as a starting point, and their collaboration
// in answering the questions needed to bring all this together. A special thank you to @Sunoo for the many hours of
// discussion and brainstorming on this and other topics.
class RtpSplitter {
    // Create an instance of RTPSplitter.
    constructor(streamingDelegate, ipFamily, serverPort, returnAudioPort, twowayAudioPort) {
        this.debug = streamingDelegate.debug;
        this.delegate = streamingDelegate;
        this.log = streamingDelegate.log;
        this.serverPort = serverPort;
        this.socket = dgram_1.createSocket(ipFamily === "ipv6" ? "udp6" : "udp4");
        // Catch errors when they happen on our splitter.
        this.socket.on("error", (error) => {
            this.log("%s: RTPSplitter Error: %s", this.delegate.protectCamera.name(), error);
            this.socket.close();
        });
        // Split the message into RTP and RTCP packets.
        this.socket.on("message", (msg) => {
            // Send RTP packets to the return audio port.
            if (this.isRtpMessage(msg)) {
                this.socket.send(msg, twowayAudioPort, "127.0.0.1");
            }
            else {
                // Save this RTCP message for heartbeat purposes for the return audio port.
                this.heartbeatMsg = Buffer.from(msg);
                // Clear the old heartbeat timer.
                clearTimeout(this.heartbeatTimer);
                this.heartbeat(twowayAudioPort);
                // RTCP control packets should go to the RTCP port.
                this.socket.send(msg, returnAudioPort, "127.0.0.1");
            }
        });
        this.debug("%s: Creating an RtpSplitter instance - inbound port: %s, twoway audio port: %s, return audio port: %s.", this.delegate.protectCamera.name(), this.serverPort, twowayAudioPort, returnAudioPort);
        // Take the socket live.
        this.socket.bind(this.serverPort);
    }
    // Send a regular heartbeat to FFmpeg to ensure the pipe remains open and the process alive.
    heartbeat(port) {
        // Clear the old heartbeat timer.
        clearTimeout(this.heartbeatTimer);
        // Send a heartbeat to FFmpeg every 3.5 seconds to keep things open. FFmpeg has a five-second timeout
        // in reading input, and we want to be comfortably within the margin for error to ensure the process
        // continues to run.
        this.heartbeatTimer = setTimeout(() => {
            this.debug("Sending ffmpeg a heartbeat.");
            this.socket.send(this.heartbeatMsg, port, "127.0.0.1");
            this.heartbeat(port);
        }, settings_1.PROTECT_TWOWAY_HEARTBEAT_INTERVAL * 1000);
    }
    // Close the socket and cleanup.
    close() {
        this.debug("%s: Closing the RtpSplitter instance on port %s.", this.delegate.protectCamera.name(), this.serverPort);
        clearTimeout(this.heartbeatTimer);
        this.socket.close();
    }
    // Retrieve the payload information from a packet to discern what the packet payload is.
    getPayloadType(message) {
        return message.readUInt8(1) & 0x7f;
    }
    // Return whether or not a packet is RTP (or not).
    isRtpMessage(message) {
        const payloadType = this.getPayloadType(message);
        return (payloadType > 90) || (payloadType === 0);
    }
}
exports.RtpSplitter = RtpSplitter;
// RTP-related utilities.
class RtpUtils {
    // Reserve consecutive ports for use with FFmpeg. FFmpeg currently lacks the ability to specify both the RTP
    // and RTCP ports. It always assumes, by convention, that when you specify an RTP port, the RTCP port is the
    // RTP port + 1. In order to work around that challenge, we need to always ensure that when we reserve multiple
    // ports for RTP (primarily for two-way audio) that we we are reserving consecutive ports only.
    static async reservePorts(count = 1) {
        // Get the first port.
        const port = await get_port_1.default();
        const ports = [port];
        // If we're requesting additional consecutive ports, keep searching until they're found.
        for (let i = 1; i < count; i++) {
            const targetConsecutivePort = port + i;
            // We need to await here in order to determine whether we need to keep going.
            // eslint-disable-next-line no-await-in-loop
            const openPort = await get_port_1.default({ port: targetConsecutivePort });
            // Unable to reserve the next consecutive port. Roll the dice again and hope for the best.
            if (openPort !== targetConsecutivePort) {
                return this.reservePorts(count);
            }
            ports.push(openPort);
        }
        return ports;
    }
}
exports.RtpUtils = RtpUtils;
//# sourceMappingURL=protect-rtp.js.map