"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProtectStreamingDelegate = void 0;
/* Copyright(C) 2017-2020, HJD (https://github.com/hjdhjd). All rights reserved.
 *
 * protect-stream.ts: Homebridge camera streaming delegate implementation for Protect.
 *
 * This module is heavily inspired by the homebridge and homebridge-camera-ffmpeg source code and
 * borrows heavily from both. Thank you for your contributions to the HomeKit world.
 */
const ffmpeg_for_homebridge_1 = __importDefault(require("ffmpeg-for-homebridge"));
const protect_ffmpeg_1 = require("./protect-ffmpeg");
const protect_rtp_1 = require("./protect-rtp");
const settings_1 = require("./settings");
// Increase the listener limits to support Protect installations with more than 10 cameras. 100 seems like a reasonable default.
// eslint-disable-next-line @typescript-eslint/no-var-requires, @typescript-eslint/no-unsafe-member-access
require("events").EventEmitter.defaultMaxListeners = 100;
// Camera streaming delegate implementation for Protect.
class ProtectStreamingDelegate {
    constructor(camera) {
        this.api = camera.api;
        this.protectCamera = camera;
        this.config = camera.platform.config;
        this.debug = camera.debug.bind(camera.platform);
        this.hap = camera.api.hap;
        this.log = camera.platform.log;
        this.name = camera.accessory.displayName;
        this.ongoingSessions = {};
        this.pendingSessions = {};
        this.platform = camera.platform;
        this.videoProcessor = this.config.videoProcessor || ffmpeg_for_homebridge_1.default || "ffmpeg";
        // Setup for our camera controller.
        const options = {
            cameraStreamCount: 2,
            delegate: this,
            streamingOptions: {
                supportedCryptoSuites: [0 /* AES_CM_128_HMAC_SHA1_80 */],
                video: {
                    resolutions: [
                        // Width, height, framerate.
                        [1920, 1080, 30],
                        [1280, 960, 30],
                        [1280, 720, 30],
                        [1024, 768, 30],
                        [640, 480, 30],
                        [640, 360, 30],
                        [480, 360, 30],
                        [480, 270, 30],
                        [320, 240, 30],
                        [320, 240, 15],
                        [320, 180, 30]
                    ],
                    codec: {
                        profiles: [0 /* BASELINE */, 1 /* MAIN */, 2 /* HIGH */],
                        levels: [0 /* LEVEL3_1 */, 1 /* LEVEL3_2 */, 2 /* LEVEL4_0 */]
                    }
                },
                audio: {
                    codecs: [
                        {
                            type: "AAC-eld" /* AAC_ELD */,
                            samplerate: 16 /* KHZ_16 */
                        }
                    ],
                    twoWayAudio: this.protectCamera.twoWayAudio
                }
            }
        };
        this.controller = new this.hap.CameraController(options);
    }
    // HomeKit image snapshot request handler.
    async handleSnapshotRequest(request, callback) {
        var _a;
        const snapshot = await this.getSnapshot(request);
        if (!snapshot) {
            callback(new Error(this.name + ": Unable to retrieve snapshot."));
            return;
        }
        // Return the image to HomeKit.
        callback(undefined, snapshot);
        // Publish the snapshot as a data URL to MQTT, if configured.
        (_a = this.protectCamera.nvr.mqtt) === null || _a === void 0 ? void 0 : _a.publish(this.protectCamera.accessory, "snapshot", "data:image/jpeg;base64," + snapshot.toString("base64"));
    }
    // Prepare to launch the video stream.
    async prepareStream(request, callback) {
        // We need to check for AAC support because it's going to inform our audio support.
        const hasLibFdk = await protect_ffmpeg_1.FfmpegProcess.codecEnabled(this.videoProcessor, "libfdk_aac");
        // Setup our audio plumbing.
        const audioReturnPort = (await protect_rtp_1.RtpUtils.reservePorts())[0];
        const audioServerPort = (hasLibFdk && this.protectCamera.twoWayAudio) ? (await protect_rtp_1.RtpUtils.reservePorts())[0] : -1;
        const audioTwoWayPort = (hasLibFdk && this.protectCamera.twoWayAudio) ? (await protect_rtp_1.RtpUtils.reservePorts(2))[0] : -1;
        const audioSSRC = this.hap.CameraController.generateSynchronisationSource();
        if (!hasLibFdk) {
            this.log("%s: Audio support disabled. A version of FFmpeg that is compiled with fdk_aac support is required to support audio.", this.protectCamera.name());
        }
        // Setup the RTP splitter for two-way audio scenarios.
        const rtpSplitter = (hasLibFdk && this.protectCamera.twoWayAudio) ?
            new protect_rtp_1.RtpSplitter(this, request.addressVersion, audioServerPort, audioReturnPort, audioTwoWayPort) : null;
        // Setup our video plumbing.
        const videoReturnPort = (await protect_rtp_1.RtpUtils.reservePorts())[0];
        const videoSSRC = this.hap.CameraController.generateSynchronisationSource();
        const sessionInfo = {
            address: request.targetAddress,
            addressVersion: request.addressVersion,
            videoPort: request.video.port,
            videoReturnPort: videoReturnPort,
            videoCryptoSuite: request.video.srtpCryptoSuite,
            videoSRTP: Buffer.concat([request.video.srtp_key, request.video.srtp_salt]),
            videoSSRC: videoSSRC,
            hasLibFdk: hasLibFdk,
            audioPort: request.audio.port,
            audioTwoWayPort: audioTwoWayPort,
            rtpSplitter: rtpSplitter,
            audioReturnPort: audioReturnPort,
            audioCryptoSuite: request.audio.srtpCryptoSuite,
            audioSRTP: Buffer.concat([request.audio.srtp_key, request.audio.srtp_salt]),
            audioSSRC: audioSSRC
        };
        // Prepare the response stream.
        const response = {
            video: {
                port: videoReturnPort,
                ssrc: videoSSRC,
                srtp_key: request.video.srtp_key,
                srtp_salt: request.video.srtp_salt
            },
            audio: {
                port: (hasLibFdk && this.protectCamera.twoWayAudio) ? audioServerPort : audioReturnPort,
                ssrc: audioSSRC,
                srtp_key: request.audio.srtp_key,
                srtp_salt: request.audio.srtp_salt
            }
        };
        // Figure out if we have the ability to deal with audio. If we do, we next have to figure out
        // if we're doing two-way audio or not. For two-way audio, we need to use a splitter to bring all
        // the pieces together. For traditional video/audio streaming, we want to keep it simple and don't
        // use a splitter.
        // Add it to the pending session queue so we're ready to start when we're called upon.
        this.pendingSessions[request.sessionID] = sessionInfo;
        callback(undefined, response);
    }
    // Launch the Protect video (and audio) stream.
    startStream(request, callback) {
        var _a, _b;
        const sessionInfo = this.pendingSessions[request.sessionID];
        // Set our packet size to be 564. Why? MPEG transport stream (TS) packets are 188 bytes in size each.
        // These packets transmit the video data that you ultimately see on your screen and are transmitted using
        // UDP. Each UDP packet is 1316 bytes in size, before being encapsulated in IP. We want to get as many
        // TS packets as we can, within reason, in those UDP packets. This translates to 1316 / 188 = 7 TS packets
        // as a limit of what can be pushed through over a network connection. Here's the problem...you need to have
        // enough data to fill that pipe, all the time. Network latency, ffmpeg overhead, and the speed / quality of
        // the original camera stream all play a role here, and as you can imagine, there's a nearly endless set of
        // combinations to deciding how to fill that pipe. Set it too low, and you're incurring extra overhead in
        // pushing less data to clients, though you're increasing interactivity by getting whatever data you have to
        // the end user. Set it too high, and startup latency becomes unacceptable when you're trying to stream.
        //
        // For audio, you have a latency problem and a packet size that's too big will force the audio to sound choppy
        // - so we opt to increase responsiveness at the risk of more overhead. This gives the end user a much better
        // audio experience, at a very marginal cost in bandwidth overhead.
        //
        // Through experimentation, I've found a sweet spot of 188 * 3 = 564 for video on Protect cameras. This works
        // very well for G3-series cameras, and pretty well for G4-series cameras. The G4s tend to push a lot more data
        // which drives the latency higher when you're first starting up a stream. In my testing, adjusting the packet
        // size beyond 564 did not have a material impact in improving the startup time of a G4 camera, but did have
        // a negative impact on G3 cameras.
        const videomtu = 188 * 3;
        const audiomtu = 188 * 1;
        // -rtsp_transport tcp: tell the RTSP stream handler that we're looking for a TCP connection.
        const fcmd = ["-hide_banner", "-rtsp_transport", "tcp", "-i", this.protectCamera.cameraUrl];
        this.log("%s: HomeKit video stream request received: %sx%s, %s fps, %s kbps.", this.protectCamera.name(), request.video.width, request.video.height, request.video.fps, request.video.max_bit_rate);
        // Configure our video parameters:
        // -map 0:v           selects the first available video track from the stream. Protect actually maps audio
        //                    and video tracks in opposite locations from where ffmpeg typically expects them. This
        //                    setting is a more general solution than naming the track locations directly in case
        //                    Protect changes this in n the future.
        // -vcodec copy       copy the stream withour reencoding it.
        // -f rawvideo        specify that we're using raw video.
        // -pix_fmt yuvj420p  use the yuvj420p pixel format, which is what Protect uses.
        // -r fps             frame rate to use for this stream. This is specified by HomeKit.
        // -b:v bitrate       the average bitrate to use for this stream. This is specified by HomeKit.
        // -bufsize size      this is the decoder buffer size, which drives the variability / quality of the output bitrate.
        // -maxrate bitrate   the maximum bitrate tolerance, used with -bufsize. We set this to max_bit_rate to effectively
        //                    create a constant bitrate.
        // -payload_type num  payload type for the RTP stream. This is negotiated by HomeKit and is usually 99 for H.264 video.
        const ffmpegVideoArgs = [
            "-map", "0:v",
            "-vcodec", "copy",
            "-f", "rawvideo",
            "-pix_fmt", "yuvj420p",
            "-r", request.video.fps.toString(),
            ...this.platform.config.ffmpegOptions.split(" "),
            "-b:v", request.video.max_bit_rate.toString() + "k",
            "-bufsize", (2 * request.video.max_bit_rate).toString() + "k",
            "-maxrate", request.video.max_bit_rate.toString() + "k",
            "-payload_type", request.video.pt.toString()
        ];
        // Add the required RTP settings and encryption for the stream:
        // -ssrc                   synchronization source stream identifier. Random number negotiated by HomeKit to identify this stream.
        // -f rtp                  specify that we're using the RTP protocol.
        // -srtp_out_suite enc     specify the output encryption encoding suites.
        // -srtp_out_params params specify the output encoding parameters. This is negotiated by HomeKit.
        const ffmpegVideoStream = [
            "-ssrc", sessionInfo.videoSSRC.toString(),
            "-f", "rtp",
            "-srtp_out_suite", "AES_CM_128_HMAC_SHA1_80",
            "-srtp_out_params", sessionInfo.videoSRTP.toString("base64"),
            "srtp://" + sessionInfo.address + ":" + sessionInfo.videoPort.toString() + "?rtcpport=" + sessionInfo.videoPort.toString() +
                "&localrtcpport=" + sessionInfo.videoPort.toString() + "&pkt_size=" + videomtu.toString()
        ];
        // Assemble the final video command line.
        fcmd.push(...ffmpegVideoArgs, ...ffmpegVideoStream);
        // Configure the audio portion of the command line, if we have a version of FFmpeg supports libfdk_aac. Options we use are:
        //
        // -map 0:a              selects the first available audio track from the stream. Protect actually maps audio
        //                       and video tracks in opposite locations from where ffmpeg typically expects them. This
        //                       setting is a more general solution than naming the track locations directly in case
        //                       Protect changes this in the future.
        // -acodec libfdk_aac    encode to AAC.
        // -profile:a aac_eld    specify enhanced, low-delay AAC for HomeKit.
        // -flags +global_header sets the global header in the bitstream.
        // -f null               null filter to pass the audio unchanged without running through a muxing operation.
        // -ar samplerate        sample rate to use for this audio. This is specified by HomeKit.
        // -b:a bitrate          bitrate to use for this audio. This is specified by HomeKit.
        // -bufsize size         this is the decoder buffer size, which drives the variability / quality of the output bitrate.
        // -ac 1                 set the number of audio channels to 1.
        // -payload_type num     payload type for the RTP stream. This is negotiated by HomeKit and is usually 110 for AAC-ELD audio.
        if (sessionInfo.hasLibFdk) {
            // Configure our video parameters.
            const ffmpegAudioArgs = [
                "-map", "0:a",
                "-acodec", "libfdk_aac",
                "-profile:a", "aac_eld",
                "-flags", "+global_header",
                "-f", "null",
                "-ar", request.audio.sample_rate.toString() + "k",
                "-b:a", request.audio.max_bit_rate.toString() + "k",
                "-bufsize", (2 * request.audio.max_bit_rate).toString() + "k",
                "-ac", "1",
                "-payload_type", request.audio.pt.toString()
            ];
            // Add the required RTP settings and encryption for the stream:
            // -ssrc                   synchronization source stream identifier. Random number negotiated by HomeKit to identify this stream.
            // -f rtp                  specify that we're using the RTP protocol.
            // -srtp_out_suite enc     specify the output encryption encoding suites.
            // -srtp_out_params params specify the output encoding parameters. This is negotiated by HomeKit.
            const ffmpegAudioStream = [
                "-ssrc", sessionInfo.audioSSRC.toString(),
                "-f", "rtp",
                "-srtp_out_suite", "AES_CM_128_HMAC_SHA1_80",
                "-srtp_out_params", sessionInfo.audioSRTP.toString("base64"),
                "srtp://" + sessionInfo.address + ":" + sessionInfo.audioPort.toString() + "?rtcpport=" + sessionInfo.audioPort.toString() +
                    "&localrtcpport=" + sessionInfo.audioPort.toString() + "&pkt_size=" + audiomtu.toString()
            ];
            fcmd.push(...ffmpegAudioArgs, ...ffmpegAudioStream);
        }
        // Additional logging, but only if we're debugging.
        if (this.platform.verboseFfmpeg) {
            fcmd.push("-loglevel", "level+verbose");
        }
        else if (this.platform.debugMode) {
            fcmd.push("-loglevel", "level+debug");
        }
        // Combine everything and start an instance of FFmpeg.
        const ffmpeg = new protect_ffmpeg_1.FfmpegProcess(this, request.sessionID, fcmd, (sessionInfo.hasLibFdk && this.protectCamera.twoWayAudio) ? undefined : { addressVersion: sessionInfo.addressVersion, port: sessionInfo.videoReturnPort }, callback);
        // Some housekeeping for our FFmpeg and splitter sessions.
        this.ongoingSessions[request.sessionID] = { ffmpeg: [ffmpeg], rtpSplitter: sessionInfo.rtpSplitter };
        delete this.pendingSessions[request.sessionID];
        // If we aren't doing two-way audio, we're done here. For two-way audio...we have some more plumbing to do.
        if (!sessionInfo.hasLibFdk || !this.protectCamera.twoWayAudio) {
            return;
        }
        const camera = this.protectCamera.accessory.context.camera;
        const sdpIpVersion = sessionInfo.addressVersion === "ipv6" ? "IP6 " : "IP4";
        // Session description protocol message that FFmpeg will share with HomeKit.
        // SDP messages tell the other side of the connection what we're expecting to receive.
        //
        // Parameters are:
        // v             protocol version - always 0.
        // o             originator and session identifier.
        // s             session description.
        // c             connection information.
        // t             timestamps for the start and end of the session.
        // m             media type - audio, adhering to RTP/AVP, payload type 110.
        // b             bandwidth information - application specific, 24k.
        // a=rtpmap      payload type 110 corresponds to an MP4 stream.
        // a=fmtp        for payload type 110, use these format parameters.
        // a=crypto      crypto suite to use for this session.
        const sdpReturnAudio = [
            "v=0",
            "o=- 0 0 IN " + sdpIpVersion + " 127.0.0.1",
            "s=" + this.name + " Audio Talkback",
            "c=IN " + sdpIpVersion + " " + sessionInfo.address,
            "t=0 0",
            "m=audio " + sessionInfo.audioTwoWayPort.toString() + " RTP/AVP 110",
            "b=AS:24",
            "a=rtpmap:110 MPEG4-GENERIC/16000/1",
            "a=fmtp:110 profile-level-id=1;mode=AAC-hbr;sizelength=13;indexlength=3;indexdeltalength=3; config=F8F0212C00BC00",
            "a=crypto:1 AES_CM_128_HMAC_SHA1_80 inline:" + sessionInfo.audioSRTP.toString("base64")
        ].join("\n");
        // Configure the audio portion of the command line, if we have a version of FFmpeg supports libfdk_aac. Options we use are:
        //
        // -protocol_whitelist   set the list of allowed protocols for this ffmpeg session.
        // -f sdp                specify that our input will be an SDP file.
        // -acodec libfdk_aac    decode AAC input.
        // -i pipe:0             read input from standard input.
        // -map 0:a              selects the first available audio track from the stream.
        // -acodec aac           encode to AAC. This is set by Protect.
        // -flags +global_header sets the global header in the bitstream.
        // -ar samplerate        sample rate to use for this audio. This is specified by Protect.
        // -b:a bitrate          bitrate to use for this audio. This is specified by Protect.
        // -ac 1                 set the number of audio channels to 1. This is specified by Protect.
        // -f adts               transmit an ADTS stream.
        const ffmpegReturnAudioCmd = [
            "-hide_banner",
            "-protocol_whitelist", "pipe,udp,rtp,file,crypto",
            "-f", "sdp",
            "-acodec", "libfdk_aac",
            "-i", "pipe:0",
            "-map", "0:a",
            "-acodec", camera.talkbackSettings.typeFmt,
            "-flags", "+global_header",
            "-ar", camera.talkbackSettings.samplingRate.toString(),
            "-b:a", "64k",
            "-ac", camera.talkbackSettings.channels.toString(),
            "-f", "adts",
            "udp://" + camera.host + ":" + camera.talkbackSettings.bindPort.toString()
        ];
        // Additional logging, but only if we're debugging.
        if (this.platform.verboseFfmpeg) {
            ffmpegReturnAudioCmd.push("-loglevel", "level+verbose");
        }
        else if (this.platform.debugMode) {
            ffmpegReturnAudioCmd.push("-loglevel", "level+debug");
        }
        const ffmpegReturnAudio = new protect_ffmpeg_1.FfmpegProcess(this, request.sessionID, ffmpegReturnAudioCmd);
        // Housekeeping for the twoway FFmpeg session.
        this.ongoingSessions[request.sessionID].ffmpeg.push(ffmpegReturnAudio);
        // Feed the SDP session description to ffmpeg on stdin.
        (_a = ffmpegReturnAudio.getStdin()) === null || _a === void 0 ? void 0 : _a.write(sdpReturnAudio);
        (_b = ffmpegReturnAudio.getStdin()) === null || _b === void 0 ? void 0 : _b.end();
    }
    // Process incoming stream requests.
    handleStreamRequest(request, callback) {
        switch (request.type) {
            case "start" /* START */:
                this.startStream(request, callback);
                break;
            case "reconfigure" /* RECONFIGURE */:
                // Once ffmpeg is updated to support this, we'll enable this one.
                this.log("%s: Ignoring request to reconfigure: %sx%s, %s fps, %s kbps.", this.protectCamera.name(), request.video.width, request.video.height, request.video.fps, request.video.max_bit_rate);
                callback();
                break;
            case "stop" /* STOP */:
            default:
                this.stopStream(request.sessionID);
                callback();
                break;
        }
    }
    // Take a snapshot.
    async getSnapshot(request) {
        const params = new URLSearchParams({ force: "true" });
        // If we have details of the snapshot request, use it to request the right size.
        if (request) {
            params.append("width", request.width.toString());
            params.append("height", request.height.toString());
        }
        // Request the image.
        const response = await this.protectCamera.nvr.nvrApi.fetch(this.protectCamera.snapshotUrl + "?" + params.toString(), { method: "GET" }, true, false);
        if (!(response === null || response === void 0 ? void 0 : response.ok)) {
            this.log("%s: Unable to retrieve snapshot.", this.protectCamera.name());
            return null;
        }
        try {
            // Process and return the image.
            return await response.buffer();
        }
        catch (error) {
            this.log.error("%s: An error occurred while making a snapshot request: %s.", this.protectCamera.name(), error);
            return null;
        }
    }
    // Close a video stream.
    stopStream(sessionId) {
        var _a, _b;
        try {
            // Stop any FFmpeg instances we have running.
            if (this.ongoingSessions[sessionId]) {
                for (const ffmpegProcess of this.ongoingSessions[sessionId].ffmpeg) {
                    ffmpegProcess.stop();
                }
            }
            // Close the splitter, if we have one.
            (_b = (_a = this.ongoingSessions[sessionId]) === null || _a === void 0 ? void 0 : _a.rtpSplitter) === null || _b === void 0 ? void 0 : _b.close();
            // Delete the entries.
            delete this.pendingSessions[sessionId];
            delete this.ongoingSessions[sessionId];
            // Inform the user.
            this.log.info("%s: Stopped video streaming session.", this.protectCamera.name());
        }
        catch (error) {
            this.log.error("%s: Error occurred while ending the FFmpeg video processes: %s.", this.protectCamera.name(), error);
        }
    }
    // Shutdown all our video streams.
    shutdown() {
        for (const session of Object.keys(this.ongoingSessions)) {
            this.stopStream(session);
        }
    }
    // Temporarily increase the verbosity of FFmpeg output for end users.
    setVerboseFfmpeg() {
        // If we're already increased our logging, we're done.
        if (this.platform.verboseFfmpeg || this.verboseFfmpegTimer) {
            return;
        }
        // Set a timer to revert back to normal behavior.
        this.verboseFfmpegTimer = setTimeout(() => {
            this.platform.verboseFfmpeg = false;
            this.log("Returning FFmpeg logging output to normal levels.");
            // Clear out the old timer.
            this.verboseFfmpegTimer = null;
        }, settings_1.PROTECT_FFMPEG_VERBOSE_DURATION * 60 * 1000);
        this.log("FFmpeg exited unexpectedly. Increasing logging output of FFmpeg for the next %s minutes to provide additional detail for future attempts to stream video.", settings_1.PROTECT_FFMPEG_VERBOSE_DURATION);
        this.platform.verboseFfmpeg = true;
    }
}
exports.ProtectStreamingDelegate = ProtectStreamingDelegate;
//# sourceMappingURL=protect-stream.js.map